"""
Utils for implementing multi-factor authentication.
"""

from datetime import timedelta
import secrets

__all__ = [
    'create_scoped_session',
    'delete_scoped_session',
    'verify_scoped_session'
]


def create_scoped_session(cache, scope, expire_after=None):
    """
    Create a scoped session (a session which has been authenticated to perform
    the given scope of work.

    The scope should describe the actor and the action, for example:

        (user, request_full_path)

    The function returns a session key that along with the scopr can be used
    to verify the session.
    """

    # Generate the key
    session_key = secrets.token_urlsafe(32)

    # Store the key
    if isinstance(expire_after, timedelta):
        expire_after = int(expire_after.total_seconds())

    cache.set(f'mfa_scoped_session:{session_key}', scope, expire_after)

    return session_key

def delete_scoped_session(cache, session_key):
    """Clear a scoped session"""
    cache.delete(f'mfa_scoped_session:{session_key}')

def verify_scoped_session(cache, session_key, scope):
    """Verify a scoped session"""
    return cache.get(f'mfa_scoped_session:{session_key}') == scope



"""
Utils for securing processes and data.
"""

import binascii
import hashlib
import random
import secrets
import time
import urllib.parse
import uuid

from . import mfa

__all__ = [
    'encrypt_password',
    'mfa',
    'random_code',
    'random_token',
    'safe_redirect_url'
]


def encrypt_password(password, salt, hash_name='sha256', iterations=100000):
    """
    Given a password and salt[1] return a hash string that cannot be reversed
    but which can be regenerated given the same password and salt.

    The name of the hash used and the number of iterations performed can be
    specified, by default the hash used is 'sha256' and the number of itertaions
    is 100,000.

    NOTE: When running tests it can be a good idea to specify a lower number of
    iterations (e.g 1) as the default 100,000 takes a while and can dramatically
    increase the amount of time it takes for tests to complete.

    NOTE: Remember the salt needs to be stored along with the password hash,
    without the salt it's not possible to test if a password from a user is
    correct.

    [1] (https://en.wikipedia.org/wiki/Salt_%28cryptography%29)
    """
    dk = hashlib.pbkdf2_hmac(
        hash_name,
        str(password).encode('utf8'),
        str(salt).encode('utf8'),
        iterations
        )
    return str(binascii.hexlify(dk))

def random_code(length=8, charset='abcdefghkmnopqrstuvw23456789'):
    """
    Generate a random code. Optionally the length of the code and the charset
    used to generate it can be specified.

    NOTE: Random codes are suitable for short term one time use, for example a
    asking a user to confirm an action by entering a code we sent them. They are
    not suitable as passwords, even if the charset is made more complex there's
    no guarantee the output code will be good/safe password choice.
    """
    rnd = random.SystemRandom()
    return ''.join([rnd.choice(charset) for i in range(0, length)])

def random_token():
    """Generate a random hexadecimal token 32 characters in length"""
    return secrets.token_urlsafe(32)

def safe_redirect_url(url, ignore_paths=None):
    """
    Return a local path that can be safely redirected to, e.g after successfully
    signing in. If a safe redirect path cannot be generated then `None` is
    returned.

    URLs are converted to a local path (e.g `'http://www.example.com/foo'`
    becomes `'/foo'`).

    Optionally a list of paths that should be ignored can be specified. It's
    common practise to specify the '/sign-out' path (or equivalent) as this can
    put the user in a never ending loop.

    NOTE: There is not check performed to see if the URL exists (e.g if it
    returns a 404.
    """

    # Parse the URL
    url_parts = urllib.parse.urlparse(url)

    # Convert the URL to a local path
    path = url_parts.path

    # Ensure the URL starts with a '/'
    if not path.startswith('/'):
        return None

    # Ensure the path starts with a single '/'
    path = '/' + path.lstrip('/')

    # Check if we should ignore this path
    ignore_paths = ignore_paths or []
    if path in ignore_paths:
        return None

    # URL is safe return it along with any query and/or fragment
    if url_parts.query:
        path += '?' + url_parts.query

    if url_parts.fragment:
        path += '#' + url_parts.fragment

    return path



import string

from manhattan import secure


def test_encrypt_password():
    """One-way encrypt a password and return the hash string"""

    password = 'abc123'
    password_salt = 'foobar'

    # Encrypting the password
    password_hash = secure.encrypt_password(password, password_salt)

    # Check the password hash is reproducable with the same details
    assert password_hash == secure.encrypt_password(password, password_salt)

    # Check that the password hash cannot be reproduced without the same salt
    assert password_hash != secure.encrypt_password(password, 'barfoo')

    # Encrypt the password using a different hash type
    password_hash = secure.encrypt_password(password, password_salt, \
            hash_name='md5')

    # Check the password hash is reproducable with the same details
    assert password_hash == secure.encrypt_password(password, password_salt, \
            hash_name='md5')

    # Check that the password hash cannot be reproduced without the same hash
    # type.
    assert password_hash != secure.encrypt_password(password, password_salt, \
            hash_name='sha1')

    # Encrypt the password using a different number of iterations
    password_hash = secure.encrypt_password(password, password_salt, \
            iterations=10)

    # Check the password hash is reproducable with the same details
    assert password_hash == secure.encrypt_password(password, password_salt, \
            iterations=10)

    # Check that the password hash cannot be reproduced without the same hash
    # type.
    assert password_hash != secure.encrypt_password(password, password_salt, \
            iterations=100)

def test_random_code():
    """Return a random code"""

    # Check we can generate a random code
    code = secure.random_code()
    assert len(code) == 8
    assert set(code).issubset('abcdefghkmnopqrstuvw23456789')

    # Check we can generate a random code of a custom length
    code = secure.random_code(length=4)
    assert len(code) == 4
    assert set(code).issubset('abcdefghkmnopqrstuvw23456789')

    # Check we can generate a random code with a custom charset
    code = secure.random_code(charset='abcd')
    assert len(code) == 8
    assert set(code).issubset('abcd')

def test_random_token():
    """Return a random token"""
    code = secure.random_token()
    assert len(code) == 32
    assert set(code).issubset(string.hexdigits)

def test_safe_redirect_url():
    """Return a URL that can safely be redirected to"""

    restricted_urls = ['/sign-out']

    # Check valid URLs are shortened to local paths
    safe_url = secure.safe_redirect_url(
            'http://www.example.com/some-path?some=query#some-fragment',
            restricted_urls
            )
    assert safe_url == '/some-path?some=query#some-fragment'

    # Check invalid URLs return None
    safe_url = secure.safe_redirect_url('www.badplace.com')
    assert safe_url == None

    # Check restricted URLs return None
    safe_url = secure.safe_redirect_url(
        'http://www.example.com/sign-out',
        restricted_urls
        )
    assert safe_url == None
